div#changelog > div.section > ul > li > p:only-child {
    margin-bottom: 0;
}

@media screen and (max-width: 768px) {
    #fork-gh img, a.github img {
        display: none;
    }
}

table.docutils {
   background-color: #fafbfc;
   border: 0;
}

table.docutils td, table.docutils th {
   border: 0;
}

table.docutils pre {
  background-color: rgba(239, 242, 244, .75);
}

pre {
  background-color: #fafbfc;
  border-left: 5px solid #558abb;
  padding: 7px 20px !important;
}

div.sidebar {
  border-top: 5px solid #558abb !important;
  background-color: #fafbfc !important;
  padding: 3px !important;
}

div.admonition,
div.caution,
div.danger,
div.error,
div.hint,
div.important,
div.note,
div.tip,
div.seealso, /* sphinx directive */
div.warning {
  background-color: #fafbfc;
  border-right: 0;
  border-top: 0;
  border-bottom: 0;
}

div.seealso {
  border-left: 5px solid #8abb55;
}

div.note {
  border-left: 5px solid #6C3082;
}

div.hint {
  border-left: 5px solid #8899a6;
}

div.tip {
  border-left: 5px solid #eef147;
}

div.caution {
  border-left: 5px solid #bb5557;
}

div.danger, div.error, div.warning {
  border-left: 5px solid #bb5557;
}

div.caution, div.danger, div.error, div.warning {
  background: rgba(160, 17, 17, 0.1);
}

code::before, code::after {
  letter-spacing: -0.2em;
  content: "\00a0";
}

code.literal {
  font-size: 75%;
  color: #24292e;
  box-sizing: border-box;
  display: inline-block;
  padding: 0;
  background: #fafcfc;
  border: 1px solid #f0f4f7;
  line-height: 18px;
}

/* BEGIN alabaster.css override for link styling */
a.reference, a.reference:hover {
  border-bottom: 0px !important;
  text-decoration-style: solid;
  text-decoration-line: underline;
}

a.reference {
  text-decoration-color: rgba(245, 230, 230, 1);
  color: rgba(249, 89, 89, 1);
}

a.reference:hover {
  text-decoration-color: rgba(250, 137, 137, 1);
  color: rgba(250, 137, 137, 1);
}
/* END alabaster.css override for link styling */

/* BEGIN alabaster.css override for fonts */

pre, tt, code {
  font-family: "Fira Mono", Consolas, Menlo, Monaco, "Courier New", Courier, monospace !important;
}

body, div.body h1, div.body h2, div.body h3, div.body h4, div.body h5, div.body h6,
div.admonition p.admonition-title, div.sphinxsidebar input, div.sphinxsidebar h3,
div.sphinxsidebar h4 {
  font-family: "Roboto", Corbel, Avenir, "Lucida Grande", "Lucida Sans", sans-serif !important;
}

body {
  font-size: 14px !important;
}
/* END alabaster.css override for fonts */

/* BEGIN alabaster.css override for logo */

div.sphinxsidebarwrapper h1.logo {
  text-align: center;
  margin-bottom: 20px;
}

div.sphinxsidebarwrapper p.blurb {
  text-align: center;
}

/* END alabaster.css override for logo styling */